#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # Timeseries
# 
# Ce notebook présente quelques étapes simples pour une série temporelle. La plupart utilise le module [statsmodels.tsa](https://www.statsmodels.org/stable/tsa.html#module-statsmodels.tsa).

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# In[2]:


get_ipython().run_line_magic('matplotlib', 'inline')


# ## Données
# 
# Les données sont artificielles mais simulent ce que pourraient être le chiffre d'affaires d'un magasin de quartier, des samedi très forts, une semaine morne, un Noël chargé, un été plat.

# In[3]:


from ensae_teaching_cs.data import generate_sells
import pandas
df = pandas.DataFrame(generate_sells())
df.head()


# ## Premiers graphiques
# 
# La série a deux saisonnalités, hebdomadaire, mensuelle.

# In[4]:


import matplotlib.pyplot as plt
fig, ax = plt.subplots(1, 2, figsize=(14, 4))
df.iloc[-30:].set_index('date').plot(ax=ax[0])
df.set_index('date').plot(ax=ax[1])
ax[0].set_title("chiffre d'affaire sur le dernier mois")
ax[1].set_title("chiffre d'affaire sur deux ans");


# Elle a une vague tendance, on peut calculer un tendance à l'ordre 1, 2, ...

# In[5]:


from statsmodels.tsa.tsatools import detrend
notrend = detrend(df.value, order=1)
df["notrend"] = notrend
df["trend"] = df['value'] - notrend
ax = df.plot(x="date", y=["value", "trend"], figsize=(14,4))
ax.set_title('tendance');


# Autocorrélations...

# In[6]:


from statsmodels.tsa.stattools import acf
cor = acf(df.value)
cor


# In[7]:


fig, ax = plt.subplots(1, 1, figsize=(14,2))
ax.plot(cor)
ax.set_title("Autocorrélogramme");


# La première saisonalité apparaît, 7, 14, 21... Les autocorrélations partielles confirment cela, plutôt 7 jours.

# In[8]:


from statsmodels.tsa.stattools import pacf
from statsmodels.graphics.tsaplots import plot_pacf
plot_pacf(df.value, lags=50);


# Comme il n'y a rien le dimanche, il vaut mieux les enlever. Garder des zéros nous priverait de modèles multiplicatifs.

# In[9]:


df["weekday"] = df.date.dt.weekday
df.head()


# In[10]:


df_nosunday = df[df.weekday != 6]
df_nosunday.head(n=10)


# In[11]:


fig, ax = plt.subplots(1, 1, figsize=(14,2))
cor = acf(df_nosunday.value)
ax.plot(cor)
ax.set_title("Autocorrélogramme");


# In[12]:


plot_pacf(df_nosunday.value, lags=50);


# On décompose la série en tendance + saisonnalité. Les étés et Noël apparaissent.

# In[13]:


from statsmodels.tsa.seasonal import seasonal_decompose
res = seasonal_decompose(df_nosunday.value, freq=7)
res.plot();


# In[14]:


plt.plot(res.seasonal[-30:])
plt.title("Saisonnalité");


# In[15]:


cor = acf(res.trend[5:-5]);
plt.plot(cor);


# On cherche maintenant la saisonnalité de la série débarrassée de sa tendance herbdomadaire. On retrouve la saisonnalité mensuelle.

# In[16]:


res_year = seasonal_decompose(res.trend[5:-5], freq=25)
res_year.plot();


# ## Test de stationnarité
# 
# Le test [KPSS](https://en.wikipedia.org/wiki/KPSS_test) permet de tester la stationnarité d'une série.

# In[17]:


from statsmodels.tsa.stattools import kpss
kpss(res.trend[5:-5])


# Comme ce n'est pas toujours facile à interpréter, on simule une variable aléatoire gaussienne donc sans tendance.

# In[18]:


from numpy.random import randn
bruit = randn(1000)
kpss(bruit)


# Et puis une série avec une tendance forte.

# In[19]:


from numpy.random import randn
from numpy import arange
bruit = randn(1000) * 100 + arange(1000) / 10
kpss(bruit)


# Une valeur forte indique une tendance et la série en a clairement une.

# ## Prédiction
# 
# Les modèles *AR*, *ARMA*, *ARIMA* se concentrent sur une série à une dimension. En machine learning, il y a la série et plein d'autres informations. On construit une matrice avec des séries décalées.

# In[20]:


from statsmodels.tsa.tsatools import lagmat
lag = 8
X = lagmat(df_nosunday["value"], lag)
lagged = df_nosunday.copy()
for c in range(1,lag+1):
    lagged["lag%d" % c] = X[:, c-1]
lagged.tail()


# On ajoute ou on réécrit le jour de la semaine qu'on utilise comme variable supplémentaire.

# In[21]:


lagged["weekday"] = lagged.date.dt.weekday


# In[22]:


X = lagged.drop(["date", "value", "notrend", "trend"], axis=1)
Y = lagged["value"]
X.shape, Y.shape


# In[23]:


from numpy import corrcoef
corrcoef(X)


# Etrange autant de grandes valeurs, cela veut dire que la tendance est trop forte pour calculer des corrélations, il vaudrait mieux tout recommencer avec la série $\Delta Y_t = Y_t - Y_{t-1}$. Bref, passons...

# In[24]:


X.columns


# Une régression linéaire car les modèles linéaires sont toujours de bonnes baseline et pour connaître le modèle simulé, on ne fera pas beaucoup mieux.

# In[25]:


from sklearn.linear_model import LinearRegression
clr = LinearRegression()
clr.fit(X, Y)


# In[26]:


from sklearn.metrics import r2_score
r2_score(Y, clr.predict(X))


# In[27]:


clr.coef_


# On retrouve la saisonnalité, $Y_t$ et $Y_{t-6}$ sont de mèches.

# In[28]:


for i in range(1, X.shape[1]):
    print("X(t-%d)" % (i), r2_score(Y, X.iloc[:, i]))


# Auparavant (l'année dernière en fait), je construisais deux bases, apprentissage et tests, comme ceci :

# In[29]:


n = X.shape[0]
X_train = X.iloc[:n * 2//3]
X_test = X.iloc[n * 2//3:]
Y_train = Y[:n * 2//3]
Y_test = Y[n * 2//3:]


# Et puis *scikit-learn* est arrivée avec [TimeSeriesSplit](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.TimeSeriesSplit.html).

# In[30]:


from sklearn.model_selection import TimeSeriesSplit
tscv = TimeSeriesSplit(n_splits=5)
for train_index, test_index in tscv.split(lagged):
    data_train, data_test = lagged.iloc[train_index, :], lagged.iloc[test_index, :]
    print("TRAIN:", data_train.shape, "TEST:", data_test.shape)


# Et on calé une forêt aléatoire...

# In[31]:


import warnings
from sklearn.ensemble import RandomForestRegressor
clr = RandomForestRegressor()

def train_test(clr, train_index, test_index):
    data_train = lagged.iloc[train_index, :]
    data_test = lagged.iloc[test_index, :]
    clr.fit(data_train.drop(["value", "date", "notrend", "trend"], 
                            axis=1), 
            data_train.value)
    r2 = r2_score(data_test.value,
                  clr.predict(data_test.drop(["value", "date", "notrend",
                                              "trend"], axis=1).values))
    return r2

warnings.simplefilter("ignore")
last_test_index = None
for train_index, test_index in tscv.split(lagged):
    r2 = train_test(clr, train_index, test_index)        
    if last_test_index is not None:
        r2_prime = train_test(clr, last_test_index, test_index)        
        print(r2, r2_prime)
    else:
        print(r2)
    last_test_index = test_index


# 2 ans coupé en 5, soit tous les 5 mois, ça veut dire que ce découpage inclut parfois Noël, parfois l'été et que les performances y seront très sensibles.

# In[32]:


from sklearn.metrics import r2_score
r2 = r2_score(data_test.value,
              clr.predict(data_test.drop(["value", "date", "notrend",
                                          "trend"], axis=1).values))
r2


# On compare avec le $r_2$ avec le même $r_2$ obtenu en utilisant $Y_{t-1}$, $Y_{t-2}$, ... $Y_{t-d}$ comme prédiction.

# In[33]:


for i in range(1, 9):
    print(i, ":", r2_score(data_test.value, data_test["lag%d" % i]))


# In[34]:


lagged[:5]


# En fait le jour de la semaine est une variable catégorielle, on crée une colonne par jour.

# In[35]:


from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import OneHotEncoder


# In[36]:


cols = ['lag1', 'lag2', 'lag3',
        'lag4', 'lag5', 'lag6', 'lag7', 'lag8']
ct = ColumnTransformer(
        [('pass', "passthrough", cols),
         ("dummies", OneHotEncoder(), ["weekday"])])
pred = ct.fit(lagged).transform(lagged[:5])
pred


# On met tout dans un pipeline parce que c'est plus joli, plus pratique aussi.

# In[37]:


from sklearn.pipeline import make_pipeline
from sklearn.decomposition import PCA, TruncatedSVD 
cols = ['lag1', 'lag2', 'lag3',
        'lag4', 'lag5', 'lag6', 'lag7', 'lag8']
model = make_pipeline(
         make_pipeline(
                 ColumnTransformer(
                    [('pass', "passthrough", cols),
                     ("dummies", make_pipeline(OneHotEncoder(), 
                                               TruncatedSVD(n_components=2)), ["weekday"])]),
        LinearRegression()))
model.fit(lagged, lagged["value"])


# C'est plus facile à voir visuellement.

# In[38]:


from mlinsights.plotting import pipeline2dot
dot = pipeline2dot(model, lagged)
from jyquickhelper import RenderJsDot
RenderJsDot(dot)


# In[39]:


r2_score(lagged['value'], model.predict(lagged))


# ## Templating
# 
# Complètement hors sujet mais utile.

# In[40]:


from jinja2 import Template
template = Template('Hello {{ name }}!')
template.render(name='John Doe')


# In[41]:


template = Template("""
{{ name }}
{{ "-" * len(name) }}
Possède :
{% for i in range(len(meubles)) %}
- {{meubles[i]}}{% endfor %}
""")
meubles = ['table', "tabouret"]
print(template.render(name='John Doe Doe', len=len,
                     meubles=meubles))


# In[42]: